// g++ -o bla bla.cpp -lppl -lgmp -O3

#include <array>

// Library for polyhedra
#include "ppl.hh"
namespace PPL = Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators; 

// Library for interval arithmetic
#include <boost/numeric/interval.hpp>
using namespace boost::numeric;
using namespace interval_lib;

// Fix rounding policies for the transcendental functions
typedef interval<double, policies<save_state<rounded_transc_std<double>>, checking_base<double>>> I;

const double min_eps=0.001; // minimum value for eps
const double min_diff_angle=0.0002; // difference with 2pi tolerated to be considered as a (quasi)corona. Sensitive!
const bool asap=true; // how L is defined for the edge potential: as small as possible (asap) if true, as large as possible (alap) otherwise
const bool verbose=false;
const I delta=I(4)*atan(I(1))/sqrt(I(12)); // candidate density

// interval variables
I r; // radius
I V111, V11r, V1r1, V1rr, Vr1r, Vrrr; // base vertex potentials (parameters : V1rr==x, V11r==y)
I E111, E11r, E1rr, Errr; // emptiness

// double variables
double m1, mr; // coefficient of angle deviation from base vertex potential
double Z1, Zr; // ceiling of vertex potential
double eps; // for epsilon-tight triangles
double l11,l1r,lrr,q11,q1r,qrr; // for edge potential

// eventual exact coronas -> additional equations on x/y to ensure theoreticaly vertex positivity
bool exists_large_corona;
bool exists_small_corona;
int large_corona[3] = {0,0,0};
int small_corona[3] = {0,0,0};

// variables for the polytopes of parameters X=V1rr, Y=V11r
PPL::Variable X(0),Y(1),M1(2),Mr(3);

// system of constraints
PPL::Constraint_System ieqs;

int nb,nt,ng; // used to count the number of different triangles checked
int type_error; // 0: polytope vide, 1: eps=0, 2: local inequality

// stupid but useful function
I val(char q)
{
    return q=='1'?I(1):r;
}

// another stupid but useful function
I Vijk(char i,char j, char k)
{
    std::string t;
    t+=i;
    t+=j;
    t+=k;    
    if (t=="111") return V111;
    if (t=="11r"||t=="r11") return V11r;
    if (t=="r1r") return Vr1r;
    if (t=="1r1") return V1r1;
    if (t=="1rr"||t=="rr1") return V1rr;
    return Vrrr;
}


#include "triangles.cpp"

#include "parameters_xy.cpp"

#include "parameters_others.cpp"

/************************************/
/* check whether HCP is the densest */
/************************************/

bool is_HCP_densest()
{
        // update values of emptiness (which depend on delta)
        E111=E(I(2),I(2),I(2),I(1),I(1),I(1));
        E11r=E(I(2),I(1)+r,I(1)+r,r,I(1),I(1));
        E1rr=E(I(2)*r,I(1)+r,I(1)+r,I(1),r,r);
        Errr=E(I(2)*r,I(2)*r,I(2)*r,r,r,r);

        // build the approximated polytope of allowed (x,y,m1,mr)
        // approximation is forced because real constraints are transcendental whereas accepted ones are rational
        // first clear all inequalities
        ieqs.clear();
        // then add constraints to have eps>=given parameter
        add_epsilon_constraints(); // augmenter paramètre par dicho tant que P non vide, sans bouger delta ?
        // finally add constraints to have positivity around any vertex
        exists_large_corona=false;
        exists_small_corona=false;
        add_vertex_positivity_constraints('1');
        add_vertex_positivity_constraints('r');
        
        // build the polytope
        PPL::C_Polyhedron P(ieqs);

        // first possible error: no suitable parameter exist!
        if (P.is_empty())
        {
            type_error=0;
            return false;
        }
        
        // set the parameters x and y
        set_xy_generic();
        if (exists_large_corona || exists_small_corona) set_xy_corona();
        
        // compute exactly m1, mr, Z1, Zr
        try
        {
            set_mZ('1');
            set_mZ('r');
        }
        catch (std::string const &err)
        {
            std::cout << err << std::endl;
            return false;
        }

        // compute exactly eps
        set_eps();
        if (eps==0)
        {
            type_error=1;
            return false;
        }

        // fix the edge potential parameters and check the local inequality for each type of triangle
        set_LQ('1','1');
        set_LQ('1','r');
        set_LQ('r','r');
        
        // check the local inequality
        try
        {
            I b=I(0,upper(I(2)*r));
            
            if (verbose) printf("Check local inequality on T111...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)+b, I(2)+b, I(2)+b, '1', '1', '1',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on T11r...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)+b, I(1)+r+b, I(1)+r+b, 'r', '1', '1',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on T1rr...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)*r+b, I(1)+r+b, I(1)+r+b, '1', 'r', 'r',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on Trrr...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)*r+b, I(2)*r+b, I(2)*r+b, 'r', 'r', 'r',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);
        }
        catch (std::string const &err)
        {
            if (verbose) std::cout << err << std::endl;
            type_error=2;
            return false;
        }
    return true;
}


/******************************************/
/* check ratios in the interval [a/d,b/d] */
/******************************************/

void check_ratios(unsigned long long int a, unsigned long long int b, unsigned long long int d, int depth)
{
    if (depth==30)
    {
        printf("Too deep recursion around r=%.30f...\n",median(r));
        return;
    }    
    r=I(a,b)/I(d);
    if (is_HCP_densest()) printf("[%llu, %llu, %llu],\n",a,b,d);
    else
    {
        printf("Refine to depth %2d ",depth+1);
        if (exists_small_corona) printf("around r-corona [%d,%d,%d] ",small_corona[0],small_corona[1],small_corona[2]);
        else if (exists_large_corona) printf("around 1-corona [%d,%d,%d] ",large_corona[0],large_corona[1],large_corona[2]);
        else printf("around r=%.30f ",median(r));
        switch(type_error)
            {
                case 0: printf("(empty parameter polytope)\n"); break;
                case 1: printf("(epsilon is too small)\n"); break;
                case 2: printf("(local inequality violated)\n");
            }

        check_ratios(2*a,a+b,2*d,depth+1);
        check_ratios(a+b,2*b,2*d,depth+1);
    }
}



int main(int argc, char *argv[])
{
    check_ratios(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),0);
    return 0;
}

